/**
 * 
 * RoadHog MapAligner
 * v0.9
 * (c) 2001 Frederik Ramm 
 * Released under GPL
 *
 * I know this needs documentation. 
 * See http://www.remote.org/frederik/roadhog/software/ for details.
 */
		 
		 package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.border.*;

import org.w3c.dom.*;
import org.apache.xerces.parsers.DOMParser;
import com.skynamics.lib.swing.SimpleGridBag;

public class MapAligner extends Gui
{

	private JMenuItem editCut;
	private JMenuItem loadPhotos;
	private JMenuItem photoLocation;

	private Map theMap;
	private Mapping[] points;

	private MapDisplay mapDisplay;

	private DOMParser parser = new DOMParser();
	private static final boolean DEBUGLINES = false;
	private File mapFile = null;

		private long totalImg = 0;
		private long totalPoint = 0;
		private long totalPaint = 0;

	public MapAligner ()
	{
		super ("RoadHog MapAligner v0.9");

		JMenuItem menuItem;
		
		setSize (800, 600);

		setJMenuBar (new JMenuBar ());
		JMenu fileMenu = new JMenu ("File");
		getJMenuBar ().add (fileMenu);

		fileMenu.add (menuItem = new JMenuItem ("Load Map..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadMap (); }
		});

		fileMenu.add (menuItem = new JMenuItem ("Save Map"));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { saveMap (); }
		});

		fileMenu.add (menuItem = new JMenuItem ("Save Map As..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { saveMapAs (); }
		});

		fileMenu.addSeparator ();

		fileMenu.add (menuItem = new JMenuItem ("Import Track..."));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { loadTrack (); }
		});

		fileMenu.addSeparator ();

		fileMenu.add (menuItem = new JMenuItem ("Quit"));
		menuItem.addActionListener (new ActionListener () { 
			public void actionPerformed (ActionEvent e) { quit (); }
		});

		JPanel mapPanel = new JPanel ();
		JPanel buttonPanel = new JPanel ();
		mapPanel.setLayout (new BorderLayout ());
		mapPanel.add (mapDisplay = new MapDisplay (), BorderLayout.CENTER);
		mapPanel.add (buttonPanel, BorderLayout.SOUTH);

		JButton resetButton = new JButton ("Recalculate");
		buttonPanel.add (resetButton);
		resetButton.addActionListener (new ActionListener () {
			public void actionPerformed (ActionEvent a) { 
				recalculateAllWeakMappings ();
				mapDisplay.repaint ();
			}
		});
		/*
		ButtonGroup bg = new ButtonGroup ();
		JToggleButton tb = new JToggleButton ("Transform");
		buttonPanel.add (tb);
		bg.add (tb);
		tb.addActionListener (new ActionListener () {
			public void actionPerformed (ActionEvent a) { repaint (); }
		});
		tb = new JToggleButton ("Grid");
		buttonPanel.add (tb);
		bg.add (tb);
		tb.addActionListener (new ActionListener () {
			public void actionPerformed (ActionEvent a) { repaint (); }
		});
		tb = new JToggleButton ("Select");
		buttonPanel.add (tb);
		bg.add (tb);
		tb.addActionListener (new ActionListener () {
			public void actionPerformed (ActionEvent a) { repaint (); }
		});
		tb.setSelected (true);
		*/

		
		Container main = getContentPane ();
        main.setLayout (new BorderLayout ());
		main.add (mapPanel, BorderLayout.CENTER);

	}

	public void loadMap ()
	{
		File loadFile = fileSelect ("XML files", new String[] { "xml", "XML" }, false);
		if (loadFile != null)	
		{
			loadMap (loadFile);
		}
	}

	public void loadMap (File mapFile)
	{
		HashMap dummy = new HashMap ();
		try
		{
			XMLUtils.readFile (parser, mapFile.getCanonicalPath (), dummy, dummy, 
				dummy, dummy, null, System.out, true, false);
		}
		catch (Exception e)
		{
			e.printStackTrace ();
		}
		if (dummy.size () == 1)
		{
			this.mapFile = mapFile;
			theMap = (Map) dummy.values ().iterator ().next ();
			mapDisplay.setMap (theMap);
			mapDisplay.setScale (20);
			Hashtable t = theMap.getCorrections ();
			Vector p = new Vector ();
			if (t != null)
			{
				for (Enumeration e = t.keys (); e.hasMoreElements (); )
				{
					Point src = (Point) e.nextElement ();
					p.add (new Mapping (src, (Point)t.get (src)));
				}
			}
			points = new Mapping[p.size ()];
			p.copyInto (points);
		}
	}

	public void saveMap ()
	{
		if (mapFile == null)
		{
			saveMapAs ();
		}
		else
		{
			saveMapAs (mapFile);
		}
	}

	public void saveMapAs ()
	{
		File saveFile = fileSelect ("XML files", new String[] { "xml", "XML" }, true);
		if (saveFile != null) 
		{
			saveMapAs (saveFile);
		}
	}

	public void saveMapAs (File saveFile)
	{

		// copy our current working data into the map object before saving

		theMap.clearCorrections ();
		
		for (int i=0; i<points.length; i++)
		{
			Mapping m = points[i];
			if (m.strong)
			{
				theMap.addCorrection (m.src, m.dst);
			}
		}

		// now save

		try 
		{
			PrintStream o = new PrintStream (new FileOutputStream (saveFile));
			o.println("<roaddb>");
			theMap.toXML (o, "   ");
			o.println("</roaddb>");
			if (o.checkError ())
				throw new IOException (null);
			o.close ();
		}
		catch (IOException ex)
		{
			errorMessage ("Error saving " + saveFile.getName(), ex);
		}
	}

	public void loadTrack ()
	{
		File loadFile = fileSelect ();
		if (loadFile != null)	
		{
			loadTrack (loadFile);
		}
	}

	public void loadTrack (File trackFile)
	{

		BufferedReader reader;

		try
		{
			reader = new BufferedReader (new FileReader (trackFile));
		}
		catch (Exception e)
		{
			errorMessage ("Cannot open file.", e);
			return;
		}

		String line;
		int ok = 0;
		int err = 0;
		Vector p = new Vector ();
		if (points != null)
		{
			for (int i=0; i<points.length; i++) p.addElement (points[i]);
		}

		try {
			while ((line = reader.readLine ()) != null)
			{

				double latitude = 0, longitude = 0;
				StringTokenizer tok = new StringTokenizer (line);
		
				if (!("TRK".equals (tok.nextToken ()))) 
					continue;
		
				String latDeg = tok.nextToken ();
				String latMin = tok.nextToken ();
				String longDeg = tok.nextToken ();
				String longMin = tok.nextToken ();
		
				try 
				{
					latitude = Integer.parseInt (latDeg.substring (1));
					latitude += Double.parseDouble (latMin) / 60.0;
					if (latDeg.startsWith("S")) latitude *= -1;
				}
				catch (NumberFormatException nfe)
				{
					continue;
				}
		
				try 
				{
					longitude = Integer.parseInt (longDeg.substring (1));
					longitude += Double.parseDouble (longMin) / 60.0;
					if (longDeg.startsWith("W")) longitude *= -1;
				}
				catch (NumberFormatException nfe)
				{
					continue;
				}

				Point2D.Double s = new Point2D.Double (longitude, latitude);	
				Point d = new Point ();
				theMap.mapGeoToPixel (s, d);
				if (mapDisplay.isOnMap (d))
				{
					Mapping m = new Mapping (d);
					if (!p.contains (m))
					{
						p.addElement (m);
						m.src = d;
					}
				}
			}
	
			reader.close ();

			if (p.isEmpty ())
			{
				points = null;
			}
			else
			{
				points = new Mapping [p.size ()];
				p.copyInto (points);
			}

			recalculateAllWeakMappings ();
			repaint ();
		}
		catch (IOException e)
		{
			errorMessage ("Error reading file.", e);
		}

	}

	public void deleteInvisiblePoints ()
	{
		Vector p = new Vector ();
		if (points != null)
		{
			for (int i=0; i<points.length; i++) p.addElement (points[i]);
		}

		for (Iterator i = p.iterator (); i.hasNext (); )
		{
			Mapping m = (Mapping) i.next ();
			if (!(m.strong || mapDisplay.isOnMap (m.dst)))
			{
				i.remove ();
			}
		}

		if (p.isEmpty ())
		{
			points = null;
		}
		else
		{
			points = new Mapping [p.size ()];
			p.copyInto (points);
		}

	}
	
	public void recalculateAllWeakMappings ()
	{

System.out.println ("recalculateAllWeakMappings");
		for (int i=0; i<points.length; i++)
		{
			if (!points[i].strong) recalculateWeakMapping (points[i]);
		}

	}

	public void recalculateWeakMappingsNear (Mapping m, int radius)
	{
		Rectangle testRect = new Rectangle (m.src.x - radius, m.src.y - radius, 2*radius, 2*radius);
		for (int i=0; i<points.length; i++)
		{
			Mapping t = points[i];
			if ((!t.strong) && testRect.contains (t.src)) recalculateWeakMapping (t);
		}
	}

	public void recalculateWeakMapping (Mapping m)
	{

		double[] c = theMap.computeCorrection (m.src);

		if ((m.deviation != null) || (c[0] != 0) || (c[1] != 0))
		{
			
			if (m.deviation == null)
			{
				m.src = new Point (m.dst);
				m.deviation = new Point ((int) (c[0] + 0.5), (int) (c[1] + 0.5));
				m.dst.translate (m.deviation.x, m.deviation.y);
			}
			else
			{
				m.deviation.setLocation ((int) (c[0] + 0.5), (int) (c[1] + 0.5));
				m.dst.setLocation (m.src.x + m.deviation.x, m.src.y + m.deviation.y);
			}
			
		}
	}

	public static void main(String args[])
	{
		MapAligner x = new MapAligner ();
		if (args.length > 0)
			x.loadMap (new File (args[0]));
		if (args.length > 1)
			x.loadTrack (new File (args[1]));
		x.setVisible (true);
	}

	public void quit ()
	{
		setVisible (false);
		System.out.println ("total in paint method: "+totalPaint);
		System.out.println ("point painting:        "+totalPoint);
		System.out.println ("image painting:        "+totalImg);
		System.exit (0);
	}

	private class MapDisplay extends JPanel
	{
		private final MapDisplayPanel mdp = new MapDisplayPanel (this);
		private final JScrollBar vsb = new JScrollBar (JScrollBar.VERTICAL);
		private final JScrollBar hsb = new JScrollBar (JScrollBar.HORIZONTAL);
		private boolean ignoreEvent = false;

		public MapDisplay ()
		{
			setLayout (new SimpleGridBag ());
			add (mdp, "x=0 y=0 wx=1 wy=1 f=b");
			add (vsb, "x=1 y=0 wx=0 wy=1 f=v");
			add (hsb, "x=0 y=1 wx=1 wy=0 f=h");

			AdjustmentListener l = new AdjustmentListener () {
				public void adjustmentValueChanged (AdjustmentEvent e) {
					if (!ignoreEvent) mdp.setOffset (hsb.getValue (), vsb.getValue ());
				}
			};

			vsb.addAdjustmentListener (l);
			hsb.addAdjustmentListener (l);
		}
		public boolean setMap (Map map) { return mdp.setMap (map); }
		public boolean isOnMap (Point t) { return mdp.isOnMap (t); }
		public void setScale (int percent) { mdp.setScale (percent); }
		public void updateScrollbars ()
		{
			ignoreEvent = true;
			Point p = mdp.getOffset ();
			Dimension ps = mdp.getPreferredSize ();
			Dimension as = mdp.getSize ();
			hsb.setValues (p.x, as.width, 0, ps.width);
			vsb.setValues (p.y, as.height, 0, ps.height);
			hsb.repaint ();
			vsb.repaint ();
			hsb.setBlockIncrement (as.width / 2);
			vsb.setBlockIncrement (as.height / 2);
			System.out.println ("mdp: pref="+ps+" actual="+as);
			System.out.println ("hsb: val="+hsb.getValue () + " ext="+hsb.getModel().getExtent()+" max="+hsb.getModel().getMaximum());
			System.out.println ("vsb: val="+vsb.getValue () + " ext="+vsb.getModel().getExtent()+" max="+vsb.getModel().getMaximum());
			ignoreEvent = false;
		}
		public void doLayout ()
		{
			super.doLayout ();
			updateScrollbars ();
		}
	}

	private class MapDisplayPanel extends JPanel 
	{

		private Mapping selectedPoint = null;
		private BasicStroke dashedStroke = new BasicStroke (1.0f, BasicStroke.CAP_SQUARE, 
			BasicStroke.JOIN_MITER, 10.0f, new float[] { 3.0f, 3.0f }, 0.0f);
		private BasicStroke wideStroke = new BasicStroke (4.0f);
		private BasicStroke plainStroke = new BasicStroke ();
		private double scale = 1;
		private Map map;
		private Point scaledPoint = new Point ();
		private Point lineStart = null;
		private Point lineEnd = null;
		private Point zoomRectOrigin = null;
		private Rectangle zoomRect = null;
		private Vector zoomHistory = new Vector ();
		private ImageIcon mapImageIcon = null;
		private Rectangle focusRect = null;
		private Image offscreenBuffer = null;
		public boolean repaintOffscreenBuffer = false;
		private MapDisplay parent;
		private Point drawOffset = new Point (0, 0);

		public MapDisplayPanel (MapDisplay p)
		{
			parent = p;
			addMouseListener (new MouseListener () {
				public void mousePressed (MouseEvent e) { mousePress (e.getPoint (),
					(e.getModifiers () & InputEvent.BUTTON3_MASK) > 0,
					(e.getModifiers () & InputEvent.BUTTON2_MASK) > 0); }
				public void mouseReleased (MouseEvent e) { mouseRelease (e.getPoint ()); }
				public void mouseClicked (MouseEvent e) { /* ignore */ }
				public void mouseEntered (MouseEvent e) { /* ignore */ }
				public void mouseExited (MouseEvent e) { mouseExit (); }
			});
			addMouseMotionListener (new MouseMotionListener () {
				public void mouseMoved (MouseEvent e) { mouseMove (e.getPoint ()); }
				public void mouseDragged (MouseEvent e) { mouseMove (e.getPoint ()); }
			});
		}

		public void setOffset (int x, int y)
		{
			if ((drawOffset.x == x) && (drawOffset.y == y)) return;
			drawOffset.setLocation (x, y);
			repaintOffscreenBuffer = true;
			repaint ();
		};
		public Point getOffset ()
		{
			return drawOffset;
		}

		public boolean setMap (Map map)
		{
			this.map = map;
			String file = (map.draftfilename != null) ? map.draftfilename : map.filename;
			mapImageIcon = new ImageIcon (file);
			try
			{
				while (mapImageIcon.getImageLoadStatus () == MediaTracker.LOADING) 
					Thread.sleep (250);
			}
			catch (InterruptedException x)
			{
				// ignore
			}

			if (mapImageIcon.getImageLoadStatus () != MediaTracker.COMPLETE)
			{
				errorMessage ("Error reading map image file.", null);
				mapImageIcon = null;
				return false;
			}

			focusRect = new Rectangle (0, 0, map.size.width, map.size.height);
			focusRect.grow (25, 25);
			deleteInvisiblePoints ();
			invalidate ();
			repaintOffscreenBuffer = true;
			repaint ();
			parent.updateScrollbars ();
			return true;
		}

		private void mousePress (Point pos, boolean rightClick, boolean midClick)
		{

			Rectangle testRect = new Rectangle ();

			if (lineStart != null)
			{
				if (rightClick)
				{
					if (lineEnd != null) repaintLine (lineStart, lineEnd);
					lineStart = null;
					lineEnd = null;
				}
				return;
			}
		
			if (points != null) for (int radius = 1; radius < 4; radius++)
			{
				testRect.setBounds ((int) ((pos.x+drawOffset.x) / scale) - radius, (int) ((pos.y+drawOffset.y) / scale) - radius, 2 * radius + 1, 2 * radius + 1);
				for (int i=0; i<points.length; i++)
				{
					Mapping m = points[i];
					if (!m.strong) continue;
					if (testRect.contains (m.dst))
					{
						if (rightClick)
						{
							System.out.println ("strong point deleted");
							// the point is no longer a strong point.
							m.strong = false;
							map.getCorrections ().remove (m.src);
							recalculateWeakMappingsNear (m, 50);
							repaintOffscreenBuffer = true;
							repaint ((int) ((m.dst.x-50) * scale) - drawOffset.x, (int) ((m.dst.y-50) * scale) - drawOffset.y, (int) (100 * scale), (int) (100 * scale));
						}
						else
						{
							selectedPoint = m;
							repaintOffscreenBuffer = true;
							repaint ((int) (m.dst.x * scale) - 5 - drawOffset.x, (int) (m.dst.y * scale) - 5 - drawOffset.y, 11, 11);
							System.out.println ("strong point move mode activated");
						}
						return;
					}
				}
	
				for (int i=0; i<points.length; i++)
				{
					Mapping m = (Mapping) points[i];
					if (m.strong) continue;
					if (testRect.contains (m.dst))
					{
						// the point is now a strong point.
						m.strong = true;
						if (m.deviation == null)
						{
							m.dst = new Point (m.src);
							m.deviation = new Point (0, 0);
						}
						map.getCorrections ().put (m.src, m.dst);
						repaintOffscreenBuffer = true;
						repaint ((int) (m.dst.x * scale) - drawOffset.x - 5, (int) (m.dst.y * scale) - drawOffset.y - 5, 11, 11);
						System.out.println ("strong point added");
						return;
					}
				}
			}

			// if we get here, user has not clicked on an existing point

			if (rightClick || midClick)
			{
				System.out.println ("zoom out");
				// zoom out
				/*
				int i = zoomHistory.size ();
				if (i > 0)
				{
					Object[] o = (Object[]) zoomHistory.elementAt (i - 1);
					scale = ((Double) o[0]).doubleValue ();
					drawOffset = (Point) o[1];
					parent.updateScrollbars ();
					zoomHistory.removeElementAt (i - 1);
				}
				else
					*/
				{
					double newscale = scale * 0.5;
					if (midClick) newscale = 1;
					System.out.println ("old draw Offset = "+drawOffset);

					drawOffset.x = (int) ((drawOffset.x / scale - getSize ().width / scale * newscale / scale) * newscale);
					drawOffset.y = (int) ((drawOffset.x / scale - getSize ().height / scale * newscale / scale) * newscale);
					if (drawOffset.x<0) drawOffset.x = 0;
					if (drawOffset.y<0) drawOffset.y = 0;
						
					scale = newscale;
					System.out.println ("new draw Offset = "+drawOffset);
					parent.updateScrollbars ();
				}
				repaintOffscreenBuffer = true;
				repaint ();
			}
			else
			{
				System.out.println ("zoom rect selection start - crop mode="+midClick);
				zoomRectOrigin = pos;
			}
		
		}
		public void mouseMove (Point pos) 
		{
			scaledPoint.setLocation ((int) ((pos.x+drawOffset.x) / scale), (int) ((pos.y+drawOffset.y) / scale));
			if (selectedPoint != null) 
			{
				if (scaledPoint.equals (selectedPoint.dst))
				{
					return;	
				}

				repaint ((int) (selectedPoint.dst.x * scale) - drawOffset.x - 5, 
					(int) (selectedPoint.dst.y * scale) - drawOffset.y - 5, 11, 11);

				selectedPoint.deviation.translate (scaledPoint.x - selectedPoint.dst.x,
					scaledPoint.y - selectedPoint.dst.y);
				selectedPoint.dst.setLocation (scaledPoint);
				recalculateWeakMappingsNear (selectedPoint, 25);
				repaintOffscreenBuffer = true;
				repaint (pos.x - (int)(30 * scale), pos.y - (int)(30 * scale), (int)(60 * scale), (int)(60 * scale));
			}
			else if (zoomRectOrigin != null)
			{
				int x = zoomRectOrigin.x;
				int width = pos.x - x;
				int y = zoomRectOrigin.y;
				int height = pos.y - y;

				if (width < 0)
				{
					x += width;
					width = -width;
				}
				if (height < 0)
				{
					y += height;
					height = -height;
				}

				if (zoomRect != null)
				{
					repaint (zoomRect.x-1, zoomRect.y-1, zoomRect.width+2, zoomRect.height+2);
					zoomRect.setBounds (x, y, width, height);
				}
				else
				{
					zoomRect = new Rectangle (x, y, width, height);
				}
				repaint (x-2, y-2, width+4, height+4);
			}
			else if (lineStart != null)
			{
				if (lineEnd == null)
				{
					lineEnd = new Point (scaledPoint);
				}
				else
				{
					repaintLine (lineStart, lineEnd);

					lineEnd.setLocation (scaledPoint);
				}
				repaintLine (lineStart, lineEnd);
			}

		}

		private void repaintLine (Point p1, Point p2)
		{
			int x1 = (int) ((p1.x < p2.x ? p1.x : p2.x)*scale)-drawOffset.x-5;
			int x2 = (int) ((p1.x > p2.x ? p1.x : p2.x)*scale)-drawOffset.x+5;
			int y1 = (int) ((p1.y < p2.y ? p1.y : p2.y)*scale)-drawOffset.y-5;
			int y2 = (int) ((p1.y > p2.y ? p1.y : p2.y)*scale)-drawOffset.y+5;
			repaint (x1, y1, x2-x1, y2-y1);
		}

		public void mouseExit () 
        {
			repaintOffscreenBuffer = (selectedPoint != null);
			zoomRectOrigin = null;
			zoomRect = null;
			lineStart = null; 
			lineEnd = null;
			selectedPoint = null;
			repaint ();
		}
	
		public void mouseRelease (Point pos) 
		{
			if (lineStart != null)
			{
				System.out.println ("division line added");
				map.addDivision (lineStart, lineEnd);
				lineStart = null;
				lineEnd = null;
			}
			else if (selectedPoint != null)
			{
				System.out.println ("strong point movement mode ended");
				recalculateWeakMappingsNear (selectedPoint, 50);
				// recalculateAllWeakMappings ();
				repaintOffscreenBuffer = true;
				repaint ();
				selectedPoint = null;
			}
			else if ((zoomRectOrigin != null) && ((zoomRect == null) || (zoomRect.width < 20) || (zoomRect.height < 20)))
			{
				lineStart = new Point ((int) ((zoomRectOrigin.x+drawOffset.x) / scale),
					(int) ((zoomRectOrigin.y+drawOffset.y) / scale));
				System.out.println ("division line creation started");
				zoomRect = null;
				zoomRectOrigin = null;
				repaint ();
				return;
			}
			else if (zoomRect != null)
			{

					/*
					System.out.println ("view crop");
					BufferedImage n = new BufferedImage ((int) (zoomRect.width / scale + 0.5), 
						(int) (zoomRect.height / scale + 0.5), BufferedImage.TYPE_INT_ARGB);
					Graphics g = n.createGraphics ();
					Point newCO = new Point ((int) (zoomRect.x / scale + 0.5), 
						(int) (zoomRect.y / scale + 0.5));

					g.drawImage (mapImage, cropOffset.x-newCO.x, cropOffset.y-newCO.y, null);
					cropOffset = newCO;
					focusRect = new Rectangle (newCO.x, newCO.y, n.getWidth (), n.getHeight ());
					focusRect.grow (25, 25);
					deleteInvisiblePoints ();
					mapImage = n;
					System.runFinalization();
					System.gc();
					*/

					//setSize (getPreferredSize ());
					//getParent ().invalidate ();
					//repaint ();

				System.out.println ("view zoom");
			
				double addScale = Math.min (getSize ().width / (double) (zoomRect.width), 
					getSize ().height / (double) (zoomRect.height));
				System.out.println ("zoom factor = "+addScale);
				/*
				zoomRect.x *= addScale;
				zoomRect.y *= addScale;
				zoomRect.width *= addScale;
				zoomRect.height *= addScale;
				*/

/*
				if (parent instanceof JViewport)
				{
					zoomRect.grow (5, 5);
					zoomHistory.addElement (new Object[] { new Double (scale), 
						new Rectangle (((JViewport)parent).getViewRect ()) });
					scale *= addScale;
					setSize (getPreferredSize ());
					invalidate ();
					parent.invalidate ();
					((JViewport) parent).scrollRectToVisible (zoomRect);
				}
				else
					*/
				{
					zoomHistory.addElement (new Object[] { new Double (scale), drawOffset });
					scale *= addScale;
					System.out.println ("zoomRect = "+zoomRect);
					System.out.println ("old draw offset = "+drawOffset);
					drawOffset.setLocation ((int)((drawOffset.x+zoomRect.x)*addScale), (int) ((drawOffset.y+zoomRect.y)*addScale));
					System.out.println ("new draw offset = "+drawOffset);
					invalidate ();
					parent.updateScrollbars ();
				}

				zoomRect = null;
				zoomRectOrigin = null;

				repaintOffscreenBuffer = true;
				repaint ();
			}
		}

		public boolean isOnMap (Point t)
		{
			if (focusRect == null) return true;
			return focusRect.contains (t);
		}

		public void setScale (int percent)
		{
			scale = ((double) percent) / 100.0;
			setSize (getPreferredSize ());
			getParent ().invalidate ();
			repaintOffscreenBuffer = true;
			repaint ();
		}

		public void updateOffscreenBuffer (Graphics oldGraphics)
		{
			offscreenBuffer = this.createImage (getSize().width, getSize().height);
			Graphics g = offscreenBuffer.getGraphics ();

long t1=System.currentTimeMillis ();
			g.drawImage (mapImageIcon.getImage (), -drawOffset.x, -drawOffset.y, 
				(int) (map.size.width * scale + 0.5), 
				(int) (map.size.height * scale + 0.5), 
				mapImageIcon.getImageObserver ());
long t2=System.currentTimeMillis ();

long t3=System.currentTimeMillis ();
			// quitter if it's all too small
			if ((scale >= 0.5) && (points != null))
			{

				Rectangle testRect = new Rectangle ((int) (drawOffset.x / scale),
					(int) (drawOffset.y / scale), 
					(int) (getSize().width / scale),
					(int) (getSize().height / scale));

				for (int i=0; i< points.length; i++)
				{
					Mapping m = points[i];
					if (!testRect.contains (m.dst)) continue;
					if (m.deviation != null)
					{
						int oldx = (int) (m.src.x * scale)-drawOffset.x;
						int oldy = (int) (m.src.y * scale)-drawOffset.y;
						g.setColor (Color.gray);
						g.fillOval (oldx - 2, oldy - 2, 5, 5);

						// for strong mappings, paint an arrow
						if (m.strong)
						{
							int newx = (int) (m.dst.x * scale)-drawOffset.x;
							int newy = (int) (m.dst.y * scale)-drawOffset.y;
							g.drawLine (oldx, oldy, newx, newy);
							GraphUtil.fill_arrowhead ((Graphics2D) g, oldx, oldy, newx, newy, 12, 12, 0.80);
						}

					}

					if (m.strong)
					{
						g.setColor (Color.red);
						g.fillOval ((int) (m.dst.x * scale)-3-drawOffset.x, (int) (m.dst.y * scale)-3-drawOffset.y, 7, 7);
					}
					else
					{
						g.setColor (Color.black);
						g.fillOval ((int) (m.dst.x * scale)-2-drawOffset.x, (int) (m.dst.y * scale)-2-drawOffset.y, 5, 5);

						if (DEBUGLINES)
						{
							// FIXME -drawOffset
						if (m.nb[1] != null)
						{
							g.setColor (Color.blue);
							line(g,m.dst.x*scale,m.dst.y*scale,m.nb[1].dst.x*scale,m.nb[1].dst.y*scale);
							line(g,m.dst.x*scale,m.dst.y*scale,m.nb[0].dst.x*scale,m.nb[0].dst.y*scale);
						} 
						else if (m.nb[0] != null)
						{
							g.setColor (Color.green);
							line(g,m.dst.x*scale,m.dst.y*scale,m.nb[0].dst.x*scale,m.nb[0].dst.y*scale);
						}
						}
		
					}
				}

				if (selectedPoint != null)
				{
					if (selectedPoint.deviation != null)
					{
						g.setColor (Color.yellow);
						g.fillOval ((int) ((selectedPoint.src.x) * scale)-3-drawOffset.x, 
							(int) ((selectedPoint.src.y) * scale)-3-drawOffset.y, 7, 7);
					}
					g.setColor (Color.blue);
					g.fillOval ((int) (selectedPoint.dst.x * scale)-3-drawOffset.x, (int) (selectedPoint.dst.y * scale)-3-drawOffset.y, 7, 7);
				}
			}
			repaintOffscreenBuffer = false;
totalPoint += (System.currentTimeMillis ()-t3);
totalImg += (t2-t1);
		}

		public void paintComponent (Graphics g)
		{
			int lastX=0, lastY=0;
			int highlightX=0, highlightY=0;
			int selectedX=0, selectedY=0, selectedIndex=-1;
	
long t0=System.currentTimeMillis ();

			super.paintComponent (g);

			if (mapImageIcon == null) 
			{
				return;
			}

			g.setColor (Color.black);
/*
			// draw a "/ / / / / " pattern as background
			if ((cropOffset.x != 0) || (cropOffset.y != 0))
			{
				int h = (int) ((cropOffset.y + mapImage.getHeight ()) * scale + 0.5);
				int w = (int) ((cropOffset.x + mapImage.getWidth ()) * scale + 0.5);
				g.clipRect (0, 0, w, h);
				for (int x = 10; x < w + h; x += 15)
				{
					g.drawLine (x, 0, x - h, h);
				}
			}
			f.translate (cropOffset.x, cropOffset.y);
			*/

			if ((offscreenBuffer == null) || (repaintOffscreenBuffer))
			{
				updateOffscreenBuffer (g);
			}

			g.drawImage (offscreenBuffer, 0, 0, this);



/*
				(int) (cropOffset.x * scale + 0.5),
				(int) (cropOffset.y * scale + 0.5),
				(int) (mapImageIcon.getIconWidth () * scale), 
				(int) (mapImageIcon.getIconHeight () * scale), 
				mapImageIcon.getImageObserver ());
				*/

			if (zoomRect != null)
			{
				((Graphics2D)g).setStroke (dashedStroke);
				((Graphics2D)g).draw (zoomRect);
			}

			((Graphics2D)g).setStroke (wideStroke);
			g.setColor (Color.green);
			if ((lineStart != null) && (lineEnd != null))
			{
				g.drawLine ((int) (lineStart.x * scale)-drawOffset.x, (int) (lineStart.y * scale)-drawOffset.y, 
					(int) (lineEnd.x * scale)-drawOffset.x, (int) (lineEnd.y * scale)-drawOffset.y);
			}

			Vector d = map.getDivisions ();
			if (d != null)
			{
				for (int i=0; i<d.size (); i++)
				{
					Map.Line l = (Map.Line) d.elementAt (i);
					g.drawLine ((int) (l.p1.x * scale)-drawOffset.x, (int) (l.p1.y * scale)-drawOffset.y, 
						(int) (l.p2.x * scale)-drawOffset.x, (int) (l.p2.y * scale)-drawOffset.y);
				}
			}

			((Graphics2D)g).setStroke (plainStroke);
			g.setColor (Color.black);


long t4=System.currentTimeMillis ();

totalPaint += (t4-t0);


		}
		private void line (Graphics g, double a, double b, double c, double d)
		{
			g.drawLine ((int)a,(int)b,(int)c,(int)d);
		}
	
		public Dimension getPreferredSize ()
		{
			if (mapImageIcon != null)
			{
				return (new Dimension ((int)(/*cropOffset.x + */ map.size.width *scale + .5), 
					(int)(/* cropOffset.y + */ map.size.height * scale+.5)));
			}
			return (new Dimension (500, 500));
		}
	
		
	}

	private class Mapping
	{
		Mapping[] nb;
		Point dst = null;
		Point deviation = null;
		Point src = null;
		boolean strong = false;
		public Mapping (Point d) { 
			dst = d; 
			src = d;
 			if (DEBUGLINES) nb = new Mapping[2];
		}
		public Mapping (Point s, Point d) {
			dst = d; 
			src = s;
			if (!s.equals (d))
			{
				deviation = new Point (d.x-s.x, d.y-s.y);
			}
 			if (DEBUGLINES) nb = new Mapping[2];
			strong = true;
		}
		public int hashCode () { return src.hashCode (); }
		public boolean equals (Object o) { return ((o instanceof Mapping) && (((Mapping)o).src.equals (src))); }
	}
}

