
package org.remote.roadhog;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.beans.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;

import org.w3c.dom.*;

import org.apache.xerces.parsers.DOMParser;

class Vertex extends XMLObject
{
	public String id;
	public String name;	
	private HashMap detail = new HashMap ();
	private HashMap description = new HashMap ();
	public String[] connector;
	public boolean passthrough[][];
	public int[] direction;
	public Edge[] edge;
	public String[] signpost;
	public int weight = 1;
	public Vector illustrations = null;
	public Vector hyperlinks = null;
	
	public Vertex ()
	{
		/* default */
	}

	public Vertex (Node node)
	throws ParseException
	{
		NodeList sublist;
		Element vertex = (Element) node;
		int length;

		sublist = vertex.getElementsByTagName ("id");
		id = getTextChild (sublist.item(0));

		sublist = vertex.getElementsByTagName ("name");
		name = (sublist.getLength () == 0) ? id : getTextChild (sublist.item(0));

		parseAttributeWithLanguage (vertex, "detail", detail);
		parseAttributeWithLanguage (vertex, "description", description);

		sublist = vertex.getElementsByTagName ("weight");
		if (sublist.getLength () != 0) weight = Integer.parseInt (getTextChild (sublist.item (0)));

		sublist = vertex.getElementsByTagName ("connector");
		length = sublist.getLength ();
		
		connector = new String[length];
		direction = new int[length];
		signpost = new String[length];
		edge = new Edge[length];
		passthrough = new boolean[length][length];
		for (int i=0; i<length; i++)
		{
			NodeList subSublist;
			Element conn = (Element) sublist.item (i);

			subSublist = conn.getElementsByTagName ("direction");
			try
			{
				direction[i] = Integer.parseInt (getTextChild (subSublist.item (0)));
			}
			catch (NumberFormatException e) 
			{ 
				/* doesn't happen due to DTD */ 
				throw new ParseException ("Bad number format in direction, vertex '" + id + "'");
			}
			
			subSublist = conn.getElementsByTagName ("id");
			connector[i] = getTextChild (subSublist.item(0));

			subSublist = conn.getElementsByTagName ("signpost");
			if (subSublist.getLength () > 0) signpost[i] = getTextChild (subSublist.item(0));
		}

		sublist = vertex.getElementsByTagName ("illustration");
		length = sublist.getLength ();
		for (int i=0; i<length; i++)
		{
			if (illustrations == null) illustrations = new Vector ();
			illustrations.addElement (new Illustration (sublist.item (i)));
		}
		sublist = vertex.getElementsByTagName ("hyperlink");
		length = sublist.getLength ();
		for (int i=0; i<length; i++)
		{
			if (hyperlinks == null) hyperlinks = new Vector ();
			hyperlinks.addElement (new Hyperlink (sublist.item (i)));
		}

		sublist = vertex.getElementsByTagName ("passthrough");
		length = sublist.getLength ();
		for (int i=0; i<length; i++)
		{
			NodeList subSublist;
			subSublist = ((Element)sublist.item(i)).getElementsByTagName ("id");
			if (subSublist.getLength () != 2)
			{
				throw new ParseException ("Passthrough node must have two id children in vertex id '" + id + "'");
			}
			int from = findConnectorIndex (getTextChild (subSublist.item (0)));
			int to = findConnectorIndex (getTextChild (subSublist.item (1)));
			if ((from == -1) || (to == -1))
			{
				throw new ParseException ("Passthrough children must be valid connector ids in vertex id '" + id + "'");
			}
			passthrough[from][to] = true;
			passthrough[to][from] = true;
		}
			
	}

	public void toXML (PrintStream o, String indent)
	{
		o.println (indent + "<vertex>");
		o.println (indent + "   <id>" + stringToXML (id) + "</id>");
		if (!id.equals (name)) o.println (indent + "   <name>" + stringToXML (name) + "</name>");
		writeAttributeWithLanguage (o, indent + "   ", "detail", detail);
		if (weight != 1)
		{
			o.println (indent + "   <weight>" + weight + "</weight>");
		}
		writeAttributeWithLanguage (o, indent + "   ", "description", description);
		if (connector != null) for (int i=0; i<connector.length; i++)
		{
			o.println (indent + "   <connector>");
			o.println (indent + "       <id>" + stringToXML (connector[i]) + "</id>");
			o.println (indent + "       <direction>" + direction[i] + "</direction>");
			if (signpost[i] != null)
				o.println (indent + "       <signpost>" + signpost[i] + "</signpost>");
			o.println (indent + "   </connector>");
		}
		if (illustrations != null)
		{
			for (Enumeration e = illustrations.elements (); e.hasMoreElements (); )
			{
				((Illustration)e.nextElement ()).toXML (o, indent + "   ");
			}
		}
		if (hyperlinks != null)
		{
			for (Enumeration e = hyperlinks.elements (); e.hasMoreElements (); )
			{
				((Hyperlink)e.nextElement ()).toXML (o, indent + "   ");
			}
		}
		o.println (indent + "</vertex>");
	}

	public int findConnectorIndex (String what)
	{
		if (connector == null) return -1;
		for (int i=0; i<connector.length; i++)
		{
			if (connector[i].equals (what)) return i;
		}
		return -1;
	}

	public String getDetail (String lang)
	{
		return (getAttributeInLanguage (detail, lang));
	}
	public String getDescription (String lang)
	{
		return (getAttributeInLanguage (description, lang));
	}
	public boolean hasDescription ()
	{
		return (description.size () > 0);
	}

	public boolean isPassthrough (int i, String j) { return isPassthrough (i, findConnectorIndex (j)); }
	public boolean isPassthrough (String i, String j) { return isPassthrough (findConnectorIndex (i), j); }
	public boolean isPassthrough (int i, int j) 
	{
		try
		{
			return passthrough[i][j];
		}
		catch (NullPointerException x)
		{
			return false;
		}
		catch (ArrayIndexOutOfBoundsException y)
		{
			return false;
		}
	}
	
	public String getUltimateDestination (Edge incoming)
	{

		int inIndex = -1;

		if (weight >= 1)
		{
			return (name);
		}

		
		for (int i=0; i<edge.length; i++)
		{
			if (edge[i] == incoming)
			{	
				inIndex = i;
				break;
			}
		}

		if (inIndex == -1)
		{
			return name;
		}


		// if the incoming edge passes through to something by default,
		// hand the request to the vertex in question
		for (int i=0; i<edge.length; i++)
		{
			if ((i != inIndex) && (edge[i] != null) && passthrough[inIndex][i])
			{
				Vertex next = edge[i].vertex[(edge[i].vertex[0] == this) ? 1 : 0];
				return next.getUltimateDestination (edge[i]);
			}
		}

		// otherwise if our own weight is "only" 0, return our name. 
		if (weight >= 0)
		{
			return (name);
		}

		// otherwise, build a mixed string from all possible directions
		StringBuffer combined = new StringBuffer ();
		for (int i=0; i<edge.length; i++)
		{
			if ((i != inIndex) && (edge[i] != null))
			{
				Vertex next = edge[i].vertex[(edge[i].vertex[0] == this) ? 1 : 0];
				if (combined.length () > 0) combined.append (" / ");
				combined.append (next.getUltimateDestination (edge[i]));
			}
		}

		return combined.toString ();

	}

	public String getBaseName ()
	{
		StringBuffer basename = new StringBuffer (id.toLowerCase ());
		for (int i = 0; i<basename.length (); i++)
		{
			if (basename.charAt (i) == ' ') basename.deleteCharAt (i);
		}
		return (basename.toString ());
	}
}
